#ifndef KEXEC_ARCH_IA64_OPTIONS_H
#define KEXEC_ARCH_IA64_OPTIONS_H

#define OPT_ARCH_MAX   (OPT_MAX+0)

#define KEXEC_ARCH_OPTIONS \
	KEXEC_OPTIONS \

#define KEXEC_ARCH_OPT_STR KEXEC_OPT_STR ""

#endif /* KEXEC_ARCH_IA64_OPTIONS_H */
